/**
 *    @file  libhttp.c
 *   @brief  HTTP parsing and building library
 *
 *  @author  François Cerbelle (Fanfan), francois@cerbelle.net
 *
 *  @internal
 *       Created:  15/11/2024
 *      Revision:  none
 * Last modified:  2024-11-19 20:24
 *      Compiler:  gcc
 *  Organization:  Cerbelle.net
 *     Copyright:  Copyright (c) 2024, François Cerbelle
 *
 *  This source code is released for free distribution under the terms of the
 *  GNU General Public License as published by the Free Software Foundation.
 */

#include "libhttp.h"

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

typedef enum HTTPVersion_e {
    HTTPVERSION_HTTP09,
    HTTPVERSION_HTTP10,
    HTTPVERSION_HTTP11,
    HTTPVERSION_HTTP11b,
    HTTPVERSION_HTTP2,
    HTTPVERSION_HTTP3,
    HTTPVERSION_INVALID
} HTTPVersion_t;

typedef enum HTTPMethod_e {
    HTTPMETHOD_GET,
    HTTPMETHOD_HEAD,
    HTTPMETHOD_POST,
    HTTPMETHOD_PUT,
    HTTPMETHOD_DELETE,
    HTTPMETHOD_CONNECT,
    HTTPMETHOD_OPTIONS,
    HTTPMETHOD_TRACE,
    HTTPMETHOD_PATCH,
    HTTPMETHOD_INVALID
} HTTPMethod_t;

typedef struct HTTPHeaders_s {
    struct HTTPHeaders_s* next;
    char* name;
    char* value;
} HTTPHeaders_t;

typedef struct HTTP_s {
    HTTPHeaders_t* headers;
    char* body;
} HTTP_t;

HTTP_t* HTTP_new() {
    HTTP_t* retval;

    if (NULL==(retval=malloc(sizeof(struct HTTP_s)))) {
        perror("HTTP_new");
        exit(EXIT_FAILURE);
    }

    if (NULL==(retval->headers=malloc(sizeof(struct HTTPHeaders_s)))) {
        perror("HTTP_new headers initialization");
        exit(EXIT_FAILURE);
    }
    retval->headers->next = NULL;
    retval->headers->name = NULL;
    retval->headers->value = NULL;

    if (NULL==(retval->body=malloc(1))) {
        perror("HTTP_new empty body allocation");
        exit(EXIT_FAILURE);
    };
    retval->body[0]=0;

    return retval;
}

void HTTP_addheader(HTTP_t* http, const char* name, const char* value) {
    HTTPHeaders_t* headers;

    assert(http);
    assert(http->headers);
    assert(http->body);

    if (NULL==(headers=malloc(sizeof(struct HTTPHeaders_s)))) {
        perror("HTTP_addheader");
        exit(EXIT_FAILURE);
    }
    if (NULL==(headers->name=malloc(strlen(name)+1))) {
        perror("HTTP_addheader name malloc");
        exit(EXIT_FAILURE);
    }
    strcpy(headers->name,name);
    if (NULL==(headers->value=malloc(strlen(value)+1))) {
        perror("HTTP_addheader value malloc");
        exit(EXIT_FAILURE);
    }
    strcpy(headers->value,value);
    headers->next = http->headers;
    http->headers = headers;
}

char* HTTP_getheaders(const HTTP_t* http) {
    HTTPHeaders_t* headers;
    char* retval;

    assert(http);
    assert(http->headers);
    assert(http->body);

    if (NULL==(retval=malloc(1))) {
        perror("HTTP_getheaders initial malloc");
        exit(EXIT_FAILURE);
    };
    retval[0]=0;
    for (headers=http->headers; headers; headers=headers->next) {
        if (headers->name && headers->value) {
            char* newretval;
            if (NULL==(newretval=realloc(retval,strlen(retval)+strlen(headers->name)+2+strlen(headers->value)+2+1))) {
                perror("HTTP_getheaders inner loop");
                exit(EXIT_FAILURE);
            }
            retval=newretval;
            strcat(retval,headers->name);
            strcat(retval,": ");
            strcat(retval,headers->value);
            strcat(retval,"\r\n");
        }
    }
    return retval;
}

void HTTP_del(HTTP_t* http) {
    HTTPHeaders_t* headers_it;
    assert(http);
    assert(http->headers);
    assert(http->body);

    while (NULL!=(headers_it=http->headers)) {
        http->headers=http->headers->next;
        if (headers_it->name)
            free(headers_it->name);
        if (headers_it->value)
            free(headers_it->value);
        free(headers_it);
    }
    free(http->body);
    free(http);
}

char* HTTP_getrequest(const HTTPMethod_t method, const char* uri, const HTTPVersion_t version) {
    char* retval;

    assert(method<=HTTPMETHOD_INVALID);
    assert(version<=HTTPVERSION_INVALID);

    if (NULL==(retval=malloc(7+1+strlen(uri)+1+8+1))) {
        perror("HTTP_getrequest");
        exit(EXIT_FAILURE);
    }
    retval[0]=0;

    switch(method) {
    case HTTPMETHOD_GET:
        strcat(retval, "GET");
        break;;
    case HTTPMETHOD_HEAD:
    case HTTPMETHOD_POST:
    case HTTPMETHOD_PUT:
    case HTTPMETHOD_DELETE:
    case HTTPMETHOD_CONNECT:
    case HTTPMETHOD_OPTIONS:
    case HTTPMETHOD_TRACE:
    case HTTPMETHOD_PATCH:
    case HTTPMETHOD_INVALID:
    default:
        fprintf(stderr,"HTTP method not supported\n");
        exit(EXIT_FAILURE);
        break;;
    }
    strcat(retval," ");

    /* NULL URI allowed, as an empty string */
    if (uri)
        strcat(retval, uri);

    strcat(retval," HTTP/");
    switch(version) {
    case HTTPVERSION_HTTP11:
    case HTTPVERSION_HTTP11b:
        strcat(retval, "1.1");
        break;;
    case HTTPVERSION_HTTP09:
    case HTTPVERSION_HTTP10:
    case HTTPVERSION_HTTP2:
    case HTTPVERSION_HTTP3:
    case HTTPVERSION_INVALID:
    default:
        fprintf(stderr,"HTTP version not supported\n");
        exit(EXIT_FAILURE);
        break;;
    }
    strcat(retval, "\r\n");
    return retval;
}

void HTTP_setbody(HTTP_t* http, const char* body) {
    char* newbody;

    assert(http);
    assert(http->body);

    if (NULL==(newbody=realloc(http->body,strlen(body)+1))) {
        perror("HTTP_setbody");
        exit(EXIT_FAILURE);
    }

    http->body = newbody;
    strcpy(http->body, body);
}

char* HTTP_getbody(HTTP_t* http) {
    assert(http);
    assert(http->body);

    return http->body;
}
